'''
Uses the CoinMetrics API to retrieve several useful pieces of fundamental data on cryptocurrencies.
'''

import datetime
import json
from enum import Enum

from finndex.util import cryptocurrencies, dateutil, mathutil, webutil
import pandas as pd

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

COIN_METRICS_API_PREFIX = "https://community-api.coinmetrics.io/v2/"
NETWORK_METRIC_SUFFIX = "assets/{}/metricdata?metrics="

COIN_METRICS_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"

# Represents an enum containing several possible keywords which can be used with the CoinMetrics API.
class CoinMetricsData(Enum):
    BLOCK_COUNT = "BlkCnt"
    MARKET_CAP = "CapMrktCurUSD"
    PRICE_USD = "PriceUSD"
    TRANSACTION_CNT = "TxCnt"
    DAILY_ADDRESSES = "AdrActCnt"

def normalize_col(col):
   '''
   ' Linearly normalizes a column within a pandas DataFrame by dividing each entry by the maximum value in the column.
   ' Returns the resultant column.
   ' 
   ' col (Series): the column of a given pandas data frame 
   '''
   return col / (col.loc[col.idxmax()])
   
def get_coinmetrics_dates(metrics_list, start_date, end_date, currencies_list, normalize_all_time = True):
   '''
   ' Retrieves a multi-layered data frame containing a list of metrics corresponding to a list of cryptocurrencies.
   ' The outer columns of the frame represent the cryptocurrencies, and the inner columns represent the retrieved metrics.
   ' 
   ' metrics_list (list<CoinMetricsData>): the list of metrics to be retrieved
   ' start_date (datetime) - the start date, with month, day, and year provided
   ' end_date (datetime) - the end date, with month, day, and year provided
   ' currencies_list (list<Cryptocurrencies>): the list of cryptocurrencies to be retrieved
   '''
   col_index = pd.MultiIndex.from_product([currencies_list, [metric.value for metric in metrics_list]])
   return_frame = pd.DataFrame(columns = col_index)
   
   for currency in currencies_list:
     desired_metrics = COIN_METRICS_API_PREFIX + NETWORK_METRIC_SUFFIX.format(currency.value.ticker.lower())
     for keyword in metrics_list:
         desired_metrics += keyword.value + ","
     desired_metrics = desired_metrics[:-1] # remove final comma
      
     page_content = json.loads(webutil.getPageContent(desired_metrics))['metricData']
     metrics_list_retrieved = page_content['metrics']
   
     metrics_frame = pd.DataFrame(page_content['series'])
     # retrieved API data contains "values" column with list of values, so explode into individual columns
     metrics_frame = pd.concat([metrics_frame.drop('values',axis=1), metrics_frame['values'].apply(pd.Series)],axis=1)
     # individual exploded columns are named with sequential natural numbers, so rename with corresponding metric
     metrics_frame = metrics_frame.rename({i:metric for (i, metric) in enumerate(metrics_list_retrieved)}, axis=1)
     metrics_frame = metrics_frame.rename({'time':'date'}, axis=1)
     metrics_frame.index = pd.to_datetime(metrics_frame['date'])
     metrics_frame.index = metrics_frame.index.tz_localize(None)
     metrics_frame.index = metrics_frame.index.floor('d')
   
     for metric in metrics_list_retrieved:
        metrics_frame[metric] = metrics_frame.apply(lambda row: float(row[metric]) if row[metric] != None else None, axis=1)
        return_frame[currency, metric] = metrics_frame[metric]

   return_frame = return_frame.astype('float')

   # linearly normalize data between 0-1 based on maximum in non-date-filtered frame
   if normalize_all_time:
      return_frame = return_frame.apply(normalize_col)

   # fill in missing entries, filter by date
   return_frame = return_frame.interpolate().loc[(return_frame.index >= start_date) & (return_frame.index <= end_date)]

   # linearly normalize data between 0-1 based on maximum in date-filtered frame
   if not normalize_all_time:
      return_frame = return_frame.apply(normalize_col)

   return return_frame

   
'''
Retrieves a dictionary containing the CoinMetrics data across all time for a given set of statistics for a given
set of cryptocurrencies. 

Valid and retrievable statistics are listed in the CoinMetricsData class.

Returns a dictionary containing a key 'metrics' pointing to all statistics which were retrieved and a key 'series' 
which points to an list of dictionaries. Each dictionary in this list contains a timestamp (key 'time') formatted to
the microsecond and a key 'values' pointing to a list of string values. The list is ordered in the same fashion
as the 'metrics' list.
'''
def getCoinMetricsDict(currenciesList, metricsList):
    returnDict = {}

    for currency in currenciesList:
        returnDict[currency] = {}
        
        desiredMetrics = COIN_METRICS_API_PREFIX + NETWORK_METRIC_SUFFIX.format(currency.value.lower())
        for keyword in metricsList:
            desiredMetrics += keyword.value + ","
        desiredMetrics = desiredMetrics[:-1] # remove final comma
        
        loadedPageData = json.loads(webutil.getPageContent(desiredMetrics))['metricData']
        metricsListRetrieved = loadedPageData['metrics']
        dataSet = pd.DataFrame(loadedPageData['series'])
        
        for dataDict in dataSet:
            valueDict = {CoinMetricsData(metricsListRetrieved[idx]):float(value) 
                     for idx, value in enumerate(dataDict['values'])}
            returnDict[currency][datetime.datetime.strptime(dataDict['time'], 
                                                                  COIN_METRICS_TIMESTAMP_FORMAT)] = valueDict
    return returnDict

'''
Retrieves from CoinMetrics a set of metrics (from [0, 1]) of a given set of currencies (type Cryptocurrencies) 
from a given date range.
'''
def getCoinMetricsDateRange(metric, startDate, endDate, currenciesList):
    dataDict = getCoinMetricsDict(currenciesList, [metric])
    
    newDict = {}
    for currency, values in dataDict.items():
        numMetrics = len(values[list(values.keys())[0]])
        
        minVal = min(val[metric] for val in values.values()) 
        maxVal = max(val[metric] for val in values.values())
            
        newDict[currency] = {}
        
        for date, metricsDict in values.items():
            if date.date() >= startDate.date() and date.date() <= endDate.date():
                newDict[currency][date] = mathutil.map(metricsDict[metric], minVal, maxVal, 0, 1) 
                
    return newDict
    
# Plots all data available from CoinMetrics across time for a given currency.
def plotAllCoinMetricsData(currency, startDate=dateutil.getCurrentDateTime() - datetime.timedelta(days=1000), endDate=dateutil.getCurrentDateTime()):
    for dataKey in list(CoinMetricsData):
        data = getCoinMetricsDateRange(dataKey, startDate, endDate, [currency])

        timeseries.TimeSeries("{}: {}".format(str(currency), str(dataKey)), {dataKey: data})



'''
Sets up the library by starting the NLP sever and initializing the graphing engine's date format.
'''

from finndex.sentiment import nlp
import pandas as pd
from pandas.plotting import register_matplotlib_converters

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2020, Crypticko"


nlp.startServer(nlp.STANFORD_NLP_TIMEOUT, nlp.STANFORD_NLP_PORT)
pd.options.plotting.backend = "plotly"


'''
Extracts data from the Fear and Greed API and displays it instantaneously and provides a historical representation in a dictionary.
'''

import datetime
import json
import pandas as pd

from finndex.util import cryptocurrencies, dateutil, mathutil, webutil

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

FEAR_AND_GREED_ADDRESS = "https://api.alternative.me/fng/?limit=0&date_format=us"
    
MIN_FEAR_AND_GREED = 0
MAX_FEAR_AND_GREED = 100

# Uses the Fear and Greed API to extract the Fear and Greed value from any given date.
def getFearAndGreed(date):
    timestampFormatted = date.strftime(dateutil.DESIRED_DATE_FORMAT) 
    
    return getAllFearAndGreed()[timestampFormatted]
   
'''
Uses the Fear and Greed API to extract all Fear and Greed values available as a range from 0-1. Returns a dictionary with key as date
and value the Fear and Greed value on that date.
'''
def getAllFearAndGreed():
    fearAndGreedVals = webutil.getPageContent(FEAR_AND_GREED_ADDRESS)
    jsonUnpacked = json.loads(fearAndGreedVals)
    dataArr = jsonUnpacked['data']
    dataDict = {}
    for singleDay in dataArr:
        timestampFormatted = dateutil.convertTimestamp(singleDay['timestamp'], '%m-%d-%Y', dateutil.DESIRED_DATE_FORMAT)
        dataDict[timestampFormatted] = mathutil.map(int(singleDay['value']), MIN_FEAR_AND_GREED, MAX_FEAR_AND_GREED, 0, 1)
        
    return dataDict

def get_all_fg():
    '''
    ' Retrieves all Fear and Greed values available; maps the values into a range from 0-1. Returns a pandas 
    ' DataFrame object containing each daily Fear and Greed value associated with its corresponding timestamp.
    '''
    
    fg_values = json.loads(webutil.getPageContent(FEAR_AND_GREED_ADDRESS))
    
    fg_data_frame = pd.DataFrame(fg_values["data"]).drop(["time_until_update", "value_classification"],axis=1)
    fg_data_frame["timestamp"] = pd.to_datetime(fg_data_frame["timestamp"])
    fg_data_frame["value"] = fg_data_frame.apply(lambda row: mathutil.map(int(row['value']), MIN_FEAR_AND_GREED, MAX_FEAR_AND_GREED, 0, 1),axis=1)
    fg_data_frame.index = fg_data_frame["timestamp"]
    fg_data_frame.index = fg_data_frame.index.rename("date")
      
    return fg_data_frame.drop("timestamp",axis=1)

def get_fg_dates(start_date, end_date, currencies_list=[cryptocurrencies.Cryptocurrencies.BITCOIN]):
    '''
    ' Retrieves the Fear and Greed values (mapped into a range between 0 and 1) between two dates,
    ' inclusive. Although Fear and Greed is not specific to a given cryptocurrency, the returned
    ' dataframe will be associated with all the cryptocurrencies provided in currencies_list to 
    ' adhere to the standardized cryptocurency-specific plotting format. In other words,
    ' the column of values will be replicated, each under the corresponding cryptocurrency.
    ' The outer column of the returned data frame represents the retrieved cryptocurrencies while the inner columns 
    ' represent the retrieved metric (only "FearGreed", in this case).
    ' 
    ' start_date (datetime) - the start date, with month, day, and year provided
    ' end_date (datetime) - the end date, with month, day, and year provided
    ' currencies_list (list) - the list of currencies to associate with the given fear and greed values
    '''
    fg_data_frame = get_all_fg()
    fg_date_filtered = fg_data_frame.loc[(fg_data_frame.index >= start_date) & 
                                     (fg_data_frame.index <= end_date)].copy() 
    
    for currency in currencies_list:
        fg_date_filtered[currency.value] = fg_date_filtered["value"]
    fg_date_filtered = fg_date_filtered.drop("value",axis=1)
    fg_date_filtered.columns = pd.MultiIndex.from_product([currencies_list, ["FearGreed"]])

    return fg_date_filtered

'''
Uses the Fear and Greed API to extract all Fear and Greed values available in a given date range. Returns a dictionary with key as date
and value the Fear and Greed value on that date.
'''
def getFearAndGreedDateRange(startDate, endDate, currenciesList=[cryptocurrencies.Cryptocurrencies.BITCOIN]):
    fearAndGreedVals = getAllFearAndGreed()
    dataDict = {}

    for date, data in fearAndGreedVals.items():
        timestampDate = datetime.datetime.strptime(date, '%Y-%m-%d')

        if timestampDate.date() >= startDate.date() and timestampDate.date() <= endDate.date(): # only consider year, month, and day
            dataDict[timestampDate.date()] = data
        
    return dataDict

# Displays a given Fear and Greed value (0-100) in a convenient gauge format.
def displayFearAndGreedNum(val, display=True):
    return gauge.Gauge(labels=['Extreme Fear','Fear','Greed','Extreme Greed'], 
      colors=['#c80000','#c84b00','#64a000', '#00c800'], currentVal=val, minVal = 0,
                 maxVal = 1, title='Fear and Greed', displayGauge=display)

# Displays the Fear and Greed value from a given date in a convenient gauge format.
def displayFearAndGreedDate(date=dateutil.getCurrentDateTime(), display=True):
    return displayFearAndGreedNum(getFearAndGreed(date), display)


'''
Uses Stanford's NLP library to analyze and the sentiment of a block of text and provides functions to stop and start the
server.
'''

import os

import numpy
from pycorenlp import StanfordCoreNLP


__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

STANFORD_NLP_LOCATION = '~/stanford-corenlp-4.1.0' # the location of the Stanford NLP library on my computer 
                                                              # (download at https://stanfordnlp.github.io/CoreNLP)
STANFORD_NLP_TIMEOUT = 100000 # the time after which a given NLP request will be killed if not yet complete
STANFORD_NLP_PORT = 9002

MIN_SENTIMENT = 0
MAX_SENTIMENT = 4

DESIRED_ARTICLES = 10

# Starts the Stanford NLP server with a given timeout and port.
def startServer(timeout, port):
   os.popen('cd {}; java -mx5g -cp "*" edu.stanford.nlp.pipeline.StanfordCoreNLPServer -timeout {} -port {} &>/dev/null'.format(
         STANFORD_NLP_LOCATION, timeout, port)) 

def find_sentiment(text):
   '''
   ' Uses the Stanford NLP library to determine the average sentiment (sentence-wise) of a given block of text in a range from
   ' 0 (Extremely Negative) to 4 (Extremely Positive). 
   '
   ' text (str): the block of text to be analyzed
   '''
   NLP_SERVER_LOCATION = 'http://localhost:{}'.format(STANFORD_NLP_PORT)
   PROPERTIES_DICTIONARY = {'annotators': 'sentiment', 'outputFormat': 'json', 'timeout': STANFORD_NLP_TIMEOUT}
   
   nlp = StanfordCoreNLP(NLP_SERVER_LOCATION)
   result = nlp.annotate(text, properties = PROPERTIES_DICTIONARY)

   sentiments = []
   for sentenceAnalysis in result['sentences']:
      sentimentValue = float(sentenceAnalysis['sentimentValue'])
      sentiments += [sentimentValue]
   return numpy.average(sentiments)

# Displays a sentiment value (0-4) in a convenient gauge format.
def displaySentimentNum(sentimentVal):
    return gauge.Gauge(labels=['Very Negative','Negative','Neutral','Positive', 'Very Positive'], 
      colors=['#c80000','#c84b00','#646400','#64a000', '#00c800'], currentVal=sentimentVal, minVal = MIN_SENTIMENT, maxVal = MAX_SENTIMENT, title='Cryptocurrency Sentiment')

# Computes the sentiment value of a given piece of text and displays it as a gauge.
def displaySentimentTxt(text):
    return displaySentimentNum(findSentiment(text))    

# Kills the server running on a given port. 
def stopServer(port):
   exec('kill $(lsof -ti tcp:{})'.format(port))

def get_nlp_date_range(start_date, end_date, currencies_list):
   '''
   ' Retrieves the 100 most popular articles for each specified cryptocurrency in the given date range.
   ' Computes the average sentiment of the articles for each currency and returns a dataframe where each
   ' column is equal to the currency and the corresponding row to the value.
   '
   ' start_date (datetime) - the start date, with month, day, and year provided
   ' end_date (datetime) - the end date, with month, day, and year provided
   ' currencies_list (list) - the list of currencies to associate with the given fear and greed values
   '''
   currency_frame = pd.DataFrame()

   for currency in currencies_list:
      articles = newsapi.get_everything(q=currency.value.name,
                                         sources='cnn,business-insider',
                                         domains='coindesk.com',
                                         from_param=start_date.strftime("%Y-%m-%d"),
                                         to=end_date.strftime("%Y-%m-%d"),
                                         language='en',
                                         sort_by='popularity',
                                         page_size = DESIRED_ARTICLES)['articles']
      
      article_content = []
      for article in articles:
         source = article['source']['id'] if article['source']['id'] != None else article['source']['name']
         url = article['url']
         source_calls = {'business-insider': webutil.get_business_insider_text, 
                    'cnn': webutil.get_cnn_text,
                    'CoinDesk': webutil.get_coin_desk_text
                   }
      
         article_content += [source_calls[source](url)]
         
      article_frame = pd.DataFrame(article_content)
      sentiment_frame = article_frame.apply(lambda row: nlp.findSentiment(row[0]), axis=1)
      
      currency_frame[currency.value] = [sentiment_frame.mean()]
   return currency_frame



from enum import Enum

class Stock:
    def __init__(self, ticker, name):
        ''' 
        ' Creates a new stock object given ticker symbol and long-form name.
        ' 
        ' ticker (str): the ticker symbol of the stock (ex.: BTC)
        ' name (str): the long-form name of the stock (ex. Bitcoin)
        '''
        self.ticker = ticker
        self.name = name

class Cryptocurrencies(Enum):
    '''
    ' Represents an enum of several cryptocurrencies corresponding to their ticker symbols and long-form names.
    '''
    BITCOIN = Stock("BTC", "Bitcoin")
    ETHEREUM = Stock("ETH", "Ethereum")
    BITCOIN_CASH = Stock("BCH", "Bitcoin Cash")
    RIPPLE = Stock("XRP", "Ripple")
    LITECOIN = Stock("LTC", "Litecoin")
    DOGECOIN = Stock("DOGE", "Dogecoin")
    TEZOS = Stock("XTZ", "Tezos")
    KYBER_NETWORK = Stock("KNC", "Kyber Network")
    CHAINLINK = Stock("LINK", "Chainlink")
    BAT = Stock("BAT", "Basic Attention Token")



'''
Provides several utility functions for extracting data from websites (including HTML pages as well as JSON files from APIs).
'''
from contextlib import closing

from bs4 import BeautifulSoup
from requests import get
from requests.exceptions import RequestException

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"


# Extracts the content located at any URL.
def getPageContent(url):
    try: 
        with closing(get(url, stream = True)) as page:
            return page.content.decode("utf-8")
    except RequestException as e:
        print(e)
        return
    
# Parses a string representing HTML, returning the parsed result for convenient iteration.
def parseHTML(url):
    return BeautifulSoup(getPageContent(url), 'html.parser')

def get_cnn_text(url):
   '''
   ' Retrieves the body of a given CNN article, excluding the headline and any advertisements.
   '
   ' url (string): the URL where the specific article is located
   '''
   htmlParser = parseHTML(url)
   
   text = ''

   for element in htmlParser.select('div'):
      if element.has_attr('class') and 'zn-body__paragraph' in element['class']: 
         text += element.text 
   text = text.replace('"', ' ')
   return text

def get_coin_desk_text(url):
   '''
   ' Retrieves the body of a given CoinDesk article.
   '
   ' url (string): the URL where the specific article is located
   '''
   parser = parseHTML(url)

   text = ""
   for element in parser.findAll("div", {"class": "article-pharagraph"}):
      text += element.text
      
   return text

def get_business_insider_text(url):
   '''
   ' Retrieves the body of a given Business Insider article.
   '
   ' url (string): the URL where the specific article is located
   '''
   parser = parseHTML(url)
   text = ""
   paragraphs = parser.findAll("p", {"class": ""})

   for element in paragraphs:
      if element.img == None:
         text += element.text

   return text.replace("\n", "").replace("\xa0", " ")


'''
Manages a set of historical data to plot the aggregate sentiment (with variable weight) relative to other values.
'''

import datetime
import functools
from enum import Enum

import ipywidgets as widgets
import numpy as np
import pandas as pd

from finndex.fundamental import coinmetrics
from finndex.sentiment import fearandgreed, trends
from finndex.util import cryptocurrencies, dateutil, mathutil
from IPython.display import display

from scipy.stats import pearsonr

__author__ = "Finn Frankis"
__copyright__ = "Copyright 2019, Crypticko"

'''
Represents a piece of data with several values over time. 'values' is a dictionary with the date as the key
and the corresponding value on that date as the value. 'slider' is the slider modifying the weight of this reading.
'''
class HistoricalDataReading:
   def __init__(self, values, slider):
      self.name = slider.description
      self.values = values
      self.slider = slider

'''
Represents all possible values that can be plotted historically. Each value corresponds to a standard data retrieval function.
'''
class HistoricalMetricType(Enum):
   FEAR_AND_GREED = functools.partial(fearandgreed.get_fg_dates)
   TRENDS = functools.partial(trends.get_trends_dates)
   BLOCK_COUNT = functools.partial(coinmetrics.get_coinmetrics_dates, [coinmetrics.CoinMetricsData.BLOCK_COUNT])
   TRANSACTION_CNT = functools.partial(coinmetrics.get_coinmetrics_dates, [coinmetrics.CoinMetricsData.TRANSACTION_CNT])
   DAILY_ADDRESSES = functools.partial(coinmetrics.get_coinmetrics_dates, [coinmetrics.CoinMetricsData.DAILY_ADDRESSES])
   MARKET_CAP = functools.partial(coinmetrics.get_coinmetrics_dates, [coinmetrics.CoinMetricsData.MARKET_CAP])
   PRICE_USD = functools.partial(coinmetrics.get_coinmetrics_dates, [coinmetrics.CoinMetricsData.PRICE_USD])

'''
Computes and plots a set of daily historical sentiment values given a set of keywords. Weights can be modified using sliders;
if weights are provided in the 'weights' parameter, presents a static graph using those weights.
'''
class HistoricalSentimentManager:
   def __init__(self, keywords_list, currencies_list, 
                        start_date = datetime.datetime.now() - datetime.timedelta(weeks=4), 
                        end_date = datetime.datetime.now(), weights = None):
      self.keywords_list = keywords_list
      self.currencies_list = currencies_list
      self.start_date = start_date
      self.end_date = end_date
      if weights != None:
         self.weights = weights
      else:
         self.weights = [1.0 / len(keywords_list) for keyword in keywords_list] # equal weighting for all values
      
      self.historical_sentiment = None

   '''
   Computes the weighted historical sentiment with the date as the key and the historical sentiment on that date as the value.
   '''
   def get_historical_sentiment(self): 
      return_frame = self.historical_sentiment

      if return_frame is None:
         frames = [metric.value(self.start_date, self.end_date, self.currencies_list) for metric in self.keywords_list]
         combined = pd.concat(frames, axis=1)
         
         return_frame = pd.DataFrame()
         for cryptocurrency in combined.columns.levels[0]:
            interp = combined[cryptocurrency].interpolate(limit_direction='backward')
            weighted = interp.apply(lambda row: np.average(row, weights=self.weights), axis=1)
            return_frame[cryptocurrency] = weighted

      self.historical_sentiment = return_frame

      return return_frame

   def get_prices(self):
      return HistoricalMetricType.PRICE_USD.value(self.start_date, self.end_date, self.currencies_list, normalize_all_time=False)

   def get_price_correlation(self):
      historical = self.get_historical_sentiment()
      prices = self.get_prices()

      return {(currency, pearsonr(historical[currency].values[1:], prices[currency]['PriceUSD'].values)[0]) for currency in self.currencies_list}


   def get_index_sentiment(self, weights=None):
      if weights == None:
         weights = [1.0 / len(self.currencies_list) for currency in self.currencies_list]
      return self.get_historical_sentiment().apply(lambda row: np.average(row, weights=weights), axis=1)